#include <stdio.h>
#include <mpi.h>
#include <omp.h>

// Function to calculate the partial area under the curve X^2 for a given range
double calculate_partial_area(double start, double end, int num_steps, int rank, int size) {
    double width = (end - start) / num_steps;
    double partial_area = 0.0;

    // Use OpenMP parallelization within each MPI process
    #pragma omp parallel reduction(+:partial_area)
    {
        // Get the number of threads and the ID of each thread
        int num_threads = omp_get_num_threads();
        int thread_id = omp_get_thread_num();

        // Loop over the subintervals assigned to the current thread within the MPI process
        for (int i = thread_id + rank * num_threads; i < num_steps; i += size * num_threads) {
            double x = start + i * width;
            double height = x * x;
            partial_area += height * width;
        }
    }

    double total_area;

    // Use MPI_Reduce to sum up the partial areas from all MPI processes
    MPI_Reduce(&partial_area, &total_area, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);

    return total_area;
}

int main(int argc, char *argv[]) {
    MPI_Init(&argc, &argv);

    int rank, size;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    int num_steps = 1000000;  // Adjust based on the required precision

    // Measure execution time
    double start_time = MPI_Wtime();

    // Calculate the total area using MPI + OpenMP hybrid parallelization
    double total_area = calculate_partial_area(0.0, 1.0, num_steps, rank, size);

    // Stop measuring time
    double end_time = MPI_Wtime();
    double elapsed_time = end_time - start_time;

    if (rank == 0) {
        // Print the results at the root process
        printf("Total Area (MPI + OpenMP): %f\n", total_area);
        printf("Time (MPI + OpenMP): %f seconds\n", elapsed_time);
    }

    MPI_Finalize();

    return 0;
}
